//************************************************************************************
function RPMGauge(options) {
  var def={ID:null
           ,className:'rpmGauge'
           ,margin:50
           ,mainTickSize:20
           ,minorTickSize:10
           ,valueFormat:'%.0f RPM'
           ,rangeStart:200
           ,rangeEnd:3000
           ,mainDiv:200
           ,minorDiv:20
           ,startAngle:-135
           ,endAngle:135
           ,verticalText:true
           ,defaultRanges:[{start:-1,end:0,class:'rangeZero'},{start:0,end:600,class:'rangeLow'},{start:600,end:2700,class:'rangeOK'},{start:2700,end:3000,class:'rangeHigh'}]
           ,userRanges:[]
           ,valueLabelPos:{x:'center',y:'center',dx:0,dy:100}
          }; 
  if (options==null) options=def;
  CombineDefaultsToObject(options,def);
  CombineDefaultsToObject(this,options);

  if ( this.ID==null ) this.ID='RPMGauge';
  this.margin=DefaultNum(options.margin,def.margin);
  this.mainTickSize=DefaultNum(options.mainTickSize,def.mainTickSize);
  this.minorTickSize=DefaultNum(options.minorTickSize,def.minorTickSize);
  this.rangeStart=DefaultNum(options.rangeStart,def.rangeStart);
  this.rangeEnd=DefaultNum(options.rangeEnd,def.rangeEnd);
  this.mainDiv=DefaultNum(options.mainDiv,def.mainDiv);
  this.minorDiv=DefaultNum(options.minorDiv,def.minorDiv);
  this.startAngle=DefaultNum(options.startAngle,def.startAngle);
  this.endAngle=DefaultNum(options.endAngle,def.endAngle);
  
  this.Value = this.rangeStart;
}
//************************************************************************************
RPMGauge.prototype = {
    init: function (svgContainer,options) {
    var def={ID:null
            ,center:{x:null,y:null} 
            ,size:null
            }; 
    if (options==null) options=def;
    options=CombineDefaultsToObject(options,def);

    this.r=200;
    this.svgContainer=svgContainer;
    var svgSizeInfo=GetSvgSizeInfo(svgContainer);
    options.center.x=DefaultNum(options.center.x,svgSizeInfo.width/2);
    options.center.y=DefaultNum(options.center.y,svgSizeInfo.height/2);
    var defSize=Math.min(svgSizeInfo.width,svgSizeInfo.height);
    options.size=DefaultNum(options.size,defSize);
    this.scale=options.size/2/this.r;
    this.center=options.center;
    this.drawGauge(svgContainer);
  },
  //************************************************************************************
  updateValue: function (value) {
    var outOfRange=(this.rangeEnd-this.rangeStart)/50;
    if ( !isNaN(value) ) {
      var needleValue=value;
      var needleValueOld=this.Value;
      needleValue=Math.max(needleValue,this.rangeStart-outOfRange);
      needleValue=Math.min(needleValue,this.rangeEnd+outOfRange);
      needleValueOld=Math.max(needleValueOld,this.rangeStart-outOfRange);
      needleValueOld=Math.min(needleValueOld,this.rangeEnd+outOfRange);
      
      this.Select("textValue").text(sprintf(this.valueFormat,value));
      this.Select("valueLabel").attr('class',this.valueRangeClass(value));
      this.rotateAnimated('needle', -this.ValueToAngle(needleValueOld), -this.ValueToAngle(needleValue), 0, 0, 200);
      this.Value = value;
    } else {
      this.rotateAnimated('needle', -this.ValueToAngle(this.Value), -this.ValueToAngle(this.rangeStart-outOfRange), 0, 0, 200);
      this.Value=this.rangeStart;
      this.Select("textValue").text('N/A');
    }
  },
  //************************************************************************************
  SubID: function (ID) {
    return this.ID+ID;
  },
  //************************************************************************************
  Select: function (ID) {
    return d3.select("#"+this.SubID(ID));
  },
  //************************************************************************************
  drawGauge: function (svg) {
    this.scale = this.drawScale(svg);
    this.drawTicks();
    this.drawNeedle();
    this.drawValueLabel();
  },
  //************************************************************************************
  drawScale: function (svg) {
    this.rotateCenterStr=this.center.x+","+this.center.y;
    this.translateAndScaleStr="translate("+this.center.x+","+this.center.y+") scale("+this.scale+")";
    
    var gauge=svg.append('g').attr('id', this.ID).attr('class', this.className); // untransformed node
    gauge=gauge.append('g').attr("transform", "translate("+this.center.x+","+this.center.y+") scale("+this.scale+")");
    // append dummy rect for outer dimensions
    gauge.append('rect').attr('id', this.SubID("Rect")).attr('x', -this.r).attr('y', -this.r).attr('width', this.r*2).attr('height',this.r).attr("fill","none").attr("stroke","none");
    var scale = gauge.append('g').attr('id', this.SubID("Scale")).attr("transform", "rotate("+this.startAngle+" 0 0)").attr('class', 'scale');

    this.drawRanges(scale,this.defaultRanges);
    this.drawRanges(scale,this.userRanges);
    scale.append('g').attr('id', this.SubID('tickmarks'));
    scale.append('g').attr('id', this.SubID('needle')).attr('class', 'needle');
    gauge.append('g').attr('id', this.SubID("gValueText"));
    return scale;
  },

  //************************************************************************************
  drawValueLabel: function (chart) {
    var Text;
    var bbox;
    
    var gauge=this.Select("gValueText");
    var Text=gauge.append("text")
      .attr("id", this.SubID("textValue")).attr("x", 0).attr("y", 0)
//      .attr("text-anchor", "middle") //.attr("dominant-baseline", "middle")
      .attr('dy', '.3em') // IE does not support 'dominant-baseline' so use dy instead
      .attr('class', 'valueText')
      .text("i"+sprintf(this.valueFormat,this.rangeEnd)+"i");
    SetElementPosition(Text,this.valueLabelPos,"#"+this.SubID("Scale"));
    bbox=Text.node().getBBox();
    Text.text('0.0');
    var r=gauge.insert('rect','#'+this.SubID('textValue')).attr("id", this.SubID("valueLabel"))
      .attr('x', bbox.x).attr('y', bbox.y)
      .attr('rx', (bbox.height*0.15)).attr('ry', (bbox.height*0.15))
      .attr('width', bbox.width).attr('height',bbox.height)
      .attr('class', 'valueTextBox');
  },
  //************************************************************************************
  valueRangeClassForRanges: function (ranges,value) {
    var classText='';
    for (var i=0;i<ranges.length;i++) {
      if ( ranges[i].start<value && value<=ranges[i].end ) classText=ranges[i].class;
    }
    return classText;
  },
  //************************************************************************************
  valueRangeClass: function (value) {
    var classText=this.valueRangeClassForRanges(this.userRanges,value);
    if ( classText=='' ) classText=this.valueRangeClassForRanges(this.defaultRanges,value);
    if ( classText=='' ) classText='valueTextBox';
    
    return classText;
  },
  //************************************************************************************
  drawRanges: function (scale,ranges) {
    var arc = d3.svg.arc()
      .startAngle(0)
      .innerRadius(this.r - this.margin-this.mainTickSize)
      .outerRadius(this.r - this.margin);
    for (var i=0;i<ranges.length;i++) {
      arc.startAngle(this.ValueToAngleRad(Math.min(Math.max(ranges[i].start,this.rangeStart),this.rangeEnd)));
      arc.endAngle(this.ValueToAngleRad(Math.max(Math.min(ranges[i].end,this.rangeEnd),this.rangeStart)));
      scale.append('path').attr('d',arc).attr('class',ranges[i].class);
    }
  },
  //************************************************************************************
  drawNeedle: function () {
    var needle = this.Select("needle"); // Add an extra group, so we can scale needle itself.
    var needleScale=this.r/70*(this.r - this.margin - this.mainTickSize)/this.r;
    needle=needle.append('g');
    needle.append('circle').attr('cx', 0).attr('cy', 0).attr('r', 8).attr("class", "center");
    needle.append("path").attr("d","M 0  5 L -5  20   0 -70 z").attr('class', 'left');
    needle.append("path").attr("d","M 0  5 L  5  20   0 -70 z").attr('class', 'right');
    needle.append("path").attr("d","M 0  5 L -5  20   5  20 z").attr('class', 'end');
    needle.attr("transform", "scale("+needleScale+")");
  },
  //************************************************************************************
  drawTicks: function () {
    var tickmarks = this.Select("tickmarks");
    var outerR=this.r - this.margin;

    var arc = d3.svg.arc()
      .startAngle(0)
      .endAngle((this.endAngle-this.startAngle)*Math.PI/180.0)
      .innerRadius(this.r - this.margin)
      .outerRadius(this.r - this.margin);
    tickmarks.append("path").attr("id", "outerRing").attr("d",arc).attr("class", "outerRing");

    // Add dummy text to get bounding box
    var SampleText=tickmarks.append("text")
        .attr("x", 0)
        .attr("y", 0)
        .attr("text-anchor", "middle")
        .attr('dy', '0.3em')
        .attr("class", "mainText")
        .text("i"+this.rangeEnd+"i");
    var box=SampleText.node().getBBox();
    SampleText.remove();
    
    for (var i = this.rangeStart; i <= this.rangeEnd; i += this.mainDiv) {
      tickmarks.append("path")
        .attr("d", "m 0,"+(-outerR)+" l 0,"+(this.mainTickSize))
        .attr("class", "mainTick")
        .attr("transform", "rotate(" + (this.ValueToAngle(i)) + " 0 0)");
      var MainTextContainer = tickmarks.append("g")
        .attr("transform", "rotate(" + (this.ValueToAngle(i)) + " 0 0)");
      var TextCenter=-outerR-box.width/2;
      var Text=MainTextContainer.append("text").attr("id", "idMainText-" + i)
        .attr("x", 0)
        .attr("y", TextCenter)
        .attr("text-anchor", "middle")
        .attr('dy', '0.3em')
        .attr("class", "mainText")
        .text(i);
      Text.attr("transform", "rotate(" + (-this.startAngle-this.ValueToAngle(i)) + " 0 "+(TextCenter)+")");
    }
    for (var i = this.rangeStart+this.minorDiv; i < this.rangeEnd; i += this.minorDiv) {
      tickmarks.append("path")
        .attr("d", "m 0,"+(-outerR)+" l 0,"+(this.minorTickSize))
        .attr("class", "minorTick")
        .attr("transform", "rotate(" + (this.ValueToAngle(i)) + " 0 0)");
    }
  },
  //************************************************************************************
  ValueToAngle: function (value) {
    return (this.endAngle-this.startAngle)/(this.rangeEnd-this.rangeStart)*(value-this.rangeStart);
  },
  //************************************************************************************
  ValueToAngleRad: function (value) {
    return this.ValueToAngle(value)*Math.PI/180.0;
  },
  //************************************************************************************
  rotateAnimated: function (selector, angleFrom, angleTo, x, y, millis) {
    var d3g = this.Select(selector);
    var rotateCenterStr=" " + x + " " + y;
    var translateAndScaleStr="";
    angleTo=angleTo % 360.0;
    var tween = function (d, i, a) {
       var aDist=angleTo-angleFrom;
//       if ( Math.abs(aDist)>180 ) aDist=(aDist>0?aDist-360:aDist+360);
       var i = function(t) { 
         return angleFrom+aDist*t; 
       }
            return function(t) {
                return 'rotate(' + (-i(t)) + "," + rotateCenterStr + ") "+translateAndScaleStr;
            };
    }
    d3g.transition()
      .duration(this.visible ? millis : 0)
      .attrTween('transform', tween);
  }
}